"""Notebook helper namespace"""

__all__ = ["patch", "update_namespace"]

import pathlib as _pathlib

from . import importer as _importer
from .patch import patch
from .. import config as _config

_PACKAGE_ROOT_PATH = _config.find_project_path()

# Add root package to sys path & install import hook
_importer.install_hook(_PACKAGE_ROOT_PATH)


def update_namespace(namespace):
    """Update a namespace with the missing metadata required to support runtime
    imports of notebooks

    :param namespace: namespace object
    :return:
    """
    # In order for relative imports to work, we need to know the __package__ name of the
    # notebook.
    namespace["__package__"] = _importer.determine_package_name(
        _pathlib.Path.cwd(), _PACKAGE_ROOT_PATH
    )



from typing import Callable, Type, TypeVar

T = TypeVar("T")


def patch(cls: Type) -> Callable[[T], T]:
    """Decorator to monkey-patch additional methods to a class.

    At import-time, this will disappear and the source code itself will be transformed
    Inside notebooks, the implementation below will be used.

    :param cls:
    :return:
    """

    def get_name(func):
        # Fix #4 to support patching (property) descriptors
        try:
            return func.__name__
        except AttributeError:
            # Support various descriptors
            for attr in "fget", "fset", "fdel", "__func__":
                try:
                    return getattr(func, attr).__name__
                except AttributeError:
                    continue

            # Raise original exception
            raise

    def _notebook_patch_impl(func):
        setattr(cls, get_name(func), func)
        return func

    return _notebook_patch_impl



import logging
import pathlib
from concurrent import futures

from .. import config
from .. import testing

logger = logging.getLogger(__name__)

DEFAULT_IGNORE_PATTERNS = (".ipynb_checkpoints",)


def _patch_nbclient_exceptions():
    """Shim to patch nbclient exception pickling"""
    import nbclient.exceptions as exc

    # Fix https://github.com/jupyter/nbclient/issues/121
    def reduce_CellExecutionError(self):
        return type(self), (self.traceback, self.ename, self.evalue)

    exc.CellExecutionError.__reduce__ = reduce_CellExecutionError


def find_notebooks(path, ignore_patterns=None):
    """Find notebooks given by a particular path.

    If the path is a directory, yield from the result of calling find_notebooks` with
    the directory path.
    If the path is a notebook file path, yield the path directly

    :param path: path to a file or directory
    :param ignore_patterns: set of patterns to ignore during recursion
    :return:
    """
    if ignore_patterns is None:
        ignore_patterns = DEFAULT_IGNORE_PATTERNS

    if path.is_dir():
        for p in path.iterdir():
            # Ignore certain files and directories e.g. .ipynb_checkpoints
            if any(path.match(p) for p in ignore_patterns):
                continue

            yield from find_notebooks(p, ignore_patterns)

    elif path.match("*.ipynb"):
        yield path


def configure(subparsers):
    project_config = config.load_default_config(config.find_project_path())

    parser = subparsers.add_parser(
        "test",
        description="Test literary notebooks in parallel",
    )
    parser.add_argument(
        "-s",
        "--source",
        type=pathlib.Path,
        default=[
            project_config.get("source_path"),
            *project_config.get("test_paths", []),
        ],
        action="append",
        help="path to notebook or directory (recursive)",
    )
    parser.add_argument(
        "-j",
        "--jobs",
        default=project_config.get("test_processes"),
        type=int,
        help="number of parallel jobs to run",
    )
    parser.add_argument(
        "-i",
        "--ignore",
        help="glob pattern to ignore during recursion",
        action="append",
    )
    return parser


def run(args):
    paths = set(
        p.resolve() for s in args.source if s for p in find_notebooks(s, args.ignore)
    )

    with futures.ProcessPoolExecutor(
        max_workers=args.jobs, initializer=_patch_nbclient_exceptions
    ) as executor:
        tasks = [executor.submit(testing.run_notebook, p) for p in paths]
        for task in futures.as_completed(tasks):
            task.result()



import traitlets
from nbconvert import preprocessors


class LiteraryTagAllowListPreprocessor(preprocessors.Preprocessor):
    allow_cell_tags = traitlets.Set(
        traitlets.Unicode(), default_value={"export", "docstring"}
    )

    def check_cell_conditions(self, cell, resources: dict, index: int) -> bool:
        tags = cell.metadata.get("tags", [])
        return bool(self.allow_cell_tags.intersection(tags))

    def preprocess(self, nb, resources: dict):
        nb.cells = [
            self.preprocess_cell(cell, resources, i)[0]
            for i, cell in enumerate(nb.cells)
            if self.check_cell_conditions(cell, resources, i)
        ]
        return nb, resources

    def preprocess_cell(self, cell, resources: dict, index: int):
        return cell, resources



import functools
import pathlib
from typing import Any, Dict
import configparser
import toml


PROJECT_ROOT_MARKERS = ("pyproject.toml", "setup.py", "setup.cfg", ".literary-project")


@functools.lru_cache()
def find_project_path(*search_paths) -> pathlib.Path:
    """Find the root path of the literary project.

    Searches for known files / directories including
    `.git/`, `.hg/`, or `pyproject.toml`.

    :param search_paths: starting search paths
    :return:
    """
    if not search_paths:
        search_paths = [pathlib.Path.cwd()]

    visited = set()

    for search_path in search_paths:
        for path in (search_path, *search_path.parents):
            # Avoid re-visiting paths
            if path in visited:
                break
            visited.add(path)

            # Look for any Python package marker
            for marker in PROJECT_ROOT_MARKERS:
                if (path / marker).exists():
                    return path

    raise FileNotFoundError("Couldn't find project path")


def load_pyproject_config(path: pathlib.Path) -> Dict[str, Any]:
    """Load configuration from a `tool.literary` section of a pyproject.toml
    file, modifying the various parameters where appropriate.

    :param path: path to pyproject.toml file
    :return:
    """
    project_path = path.parent.absolute()
    data = toml.load(path)

    config = {
        k.replace("-", "_"): v
        for k, v in data.get("tool", {}).get("literary", {}).items()
    }

    return load_config_from_dict(project_path, config)


def load_setup_cfg_config(path: pathlib.Path) -> Dict[str, Any]:
    """Load configuration from the `literary` section of a setup.cfg file,
    modifying the various parameters where appropriate.

    :param path: path to setup.cfg file
    :return:
    """
    project_path = path.parent.absolute()

    parser = configparser.ConfigParser()
    parser.read(path)

    if parser.has_section("literary"):
        config = dict(parser.items("literary"))
    else:
        config = {}

    return load_config_from_dict(project_path, config)


def load_config_from_dict(
    project_path: pathlib.Path, config: Dict[str, Any]
) -> Dict[str, Any]:

    source_path = config.get("source_path")
    if source_path is not None:
        source_path = (project_path / source_path).resolve()

    package_path = config.get("package_path")
    if package_path is not None:
        package_path = (project_path / package_path).resolve()

    test_paths = config.get("test_paths", [])
    if test_paths:
        test_paths = [(project_path / p).resolve() for p in test_paths]

    test_processes = config.get("test_processes")

    return {
        "source_path": source_path,
        "package_path": package_path,
        "test_paths": test_paths,
        "test_processes": test_processes,
    }


def load_default_config(project_path: pathlib.Path) -> Dict[str, Any]:
    """Load configuration from a pyproject.toml file, returning an empty
    dictionary if the file does not exist

    :param project_path: path to pyproject.toml file
    :return:
    """
    pyproject_path = project_path / "pyproject.toml"
    if pyproject_path.exists():
        return load_pyproject_config(pyproject_path)

    setup_cfg_path = project_path / "setup.cfg"
    if setup_cfg_path.exists():
        return load_setup_cfg_config(setup_cfg_path)

    return {}



import pathlib

from .. import package
from .. import config


def configure(subparsers):
    project_config = config.load_default_config(config.find_project_path())

    parser = subparsers.add_parser(
        "build",
        description="Build a pure-Python package from a set of Jupyter notebooks",
    )
    parser.add_argument(
        "-s",
        "--source",
        type=pathlib.Path,
        default=project_config.get("source_path"),
        help="source directory for notebooks",
    )
    parser.add_argument(
        "-p",
        "--package",
        type=pathlib.Path,
        default=project_config.get("package_path"),
        help="destination path generated package",
    )
    parser.add_argument(
        "-i",
        "--ignore",
        help="glob pattern to ignore during recursion",
        action="append",
    )
    return parser


def run(args):
    if args.source is None:
        raise ValueError(f"Invalid source path {args.source!r}")
    if args.package is None:
        raise ValueError(f"Invalid package path {args.package!r}")

    package.build_package(args.source, args.package, args.ignore)



import pathlib
import shutil

import nbformat
import traitlets.config

from . import exporter


DEFAULT_IGNORE_PATTERNS = (".ipynb_checkpoints", "__pycache__", ".*")


def create_notebook_code(nb_exporter, path: pathlib.Path) -> str:
    """Return the source code for a given notebook file

    :param nb_exporter: nbconvert exporter instance
    :param path: path to notebook
    :return:
    """
    nb = nbformat.read(path, as_version=nbformat.NO_CONVERT)
    body, resources = nb_exporter.from_notebook_node(nb)
    return body


def build_package_component(
    nb_exporter,
    source_dir_path,
    dest_dir_path,
    ignore_patterns=None,
):
    """Recursively build a pure-Python package from a source tree

    :param nb_exporter: nbconvert exporter instance
    :param source_dir_path: path to current source directory
    :param dest_dir_path: path to current destination directory
    :param ignore_patterns: glob patterns of files and directories to ignore during
    recursion
    :return:
    """
    if ignore_patterns is None:
        ignore_patterns = DEFAULT_IGNORE_PATTERNS

    for path in source_dir_path.iterdir():
        # Ignore any unwanted files or directories
        if any(path.match(p) for p in ignore_patterns):
            continue

        relative_path = path.relative_to(source_dir_path)

        # Rewrite notebook in target directory
        if path.match("*.ipynb"):
            dest_path = dest_dir_path / relative_path.with_suffix(".py")

            source = create_notebook_code(nb_exporter, path)
            dest_path.write_text(source)

        # Recurse into directory
        elif path.is_dir():
            dest_path = dest_dir_path / relative_path
            dest_path.mkdir(parents=True, exist_ok=True)
            build_package_component(nb_exporter, path, dest_path)

        # Copy file directly
        else:
            dest_path = dest_dir_path / relative_path
            dest_path.write_bytes(
                path.read_bytes(),
            )


def build_package(
    source_path: pathlib.Path,
    dest_path: pathlib.Path,
    ignore_patterns=None,
):
    """Build a pure-Python package from a literary source tree

    :param source_path: path to notebooks directory
    :param dest_path: path to generated package
    :param ignore_patterns: glob patterns of files and directories to ignore during
    recursion
    :return:
    """
    c = traitlets.config.Config()
    nb_exporter = exporter.LiteraryPythonExporter(config=c)

    if dest_path.exists():
        shutil.rmtree(dest_path)
    dest_path.mkdir()

    build_package_component(nb_exporter, source_path, dest_path, ignore_patterns)



import importlib.machinery
import linecache
import traceback
import os
import pathlib
import sys
from typing import AnyStr, Union

import nbformat
import traitlets.config

from ..exporter import LiteraryPythonExporter


class NotebookLoader(importlib.machinery.SourcelessFileLoader):

    def __init__(self, fullname: str, path: str, config):
        super().__init__(fullname, path)

        self.config = config

    def _update_linecache(self, path: str, source: str):
        linecache.cache[path] = (
            len(source),
            None,
            source.splitlines(keepends=True),
            path,
        )

    def get_code(self, fullname: str):
        path = self.get_filename(fullname)
        body = self.get_transpiled_source(path)
        # Ensure that generated source is available for tracebacks
        self._update_linecache(path, body)
        return compile(body, path, 'exec')

    def get_transpiled_source(self, path: str):
        nb = nbformat.read(path, as_version=nbformat.NO_CONVERT)
        exporter = LiteraryPythonExporter(config=self.config)
        body, resources = exporter.from_notebook_node(nb)
        return body


class NotebookFinder(importlib.abc.MetaPathFinder):
    def __init__(self, path, config):
        self._path = path
        self._config = config

    def _load_package_spec(self, path: pathlib.Path,
                           name: str) -> importlib.machinery.ModuleSpec:
        spec = importlib.machinery.ModuleSpec(
            name,
            NotebookLoader(name, str(path), self._config),
            origin=str(path),
            is_package=True,
        )
        spec.submodule_search_locations = [str(path.parent)]
        spec.has_location = True
        return spec

    def _load_module_spec(self, path: pathlib.Path,
                          name: str) -> importlib.machinery.ModuleSpec:
        spec = importlib.machinery.ModuleSpec(
            name, NotebookLoader(name, str(path), self._config), origin=str(path)
        )
        spec.has_location = True
        return spec

    def find_spec(self, fullname, path, target=None):
        if path is None:
            path = self._path
        head, _, module_name = fullname.rpartition(".")

        for entry in path:
            # If we can find {module_name}/__init__.ipynb, this is a package
            package_path = pathlib.Path(entry) / module_name / "__init__.ipynb"
            if package_path.exists():
                return self._load_package_spec(package_path, fullname)

            # Otherwise look for the notebook itself
            module_path = pathlib.Path(entry) / f"{module_name}.ipynb"
            if module_path.exists():
                return self._load_module_spec(module_path, fullname)


def determine_package_name(path: pathlib.Path, package_root_path: pathlib.Path) -> str:
    """Determine the corresponding importable name for a package directory given by
    a particular file path

    :param path: path to package
    :param package_root_path: root path containing notebook package directory
    :return:
    """
    relative_path = path.relative_to(package_root_path)
    return ".".join(relative_path.parts)


def install_hook(
        package_root_path: Union[AnyStr, os.PathLike], set_except_hook: bool = True
):
    """Install notebook import hook

    Don't allow the user to specify a custom search path, because we also need this to
    interoperate with the default Python module importers which use sys.path

    :param package_root_path: root path containing notebook package directory
    :param set_except_hook: overwrite `sys.excepthook` to correctly display tracebacks
    inside notebooks
    :return:
    """
    # Make notebook packages importable by adding package root path to sys.path
    sys.path.append(str(package_root_path))
    sys.meta_path.insert(0, NotebookFinder(sys.path, config=traitlets.config.Config()))

    # Python's C-level traceback reporting doesn't call `linecache`, and so retrieves
    # the underlying notebook source instead of the generated Python code
    if set_except_hook:
        sys.excepthook = traceback.print_exception
